package leodbc.javaagent.generation;

import java.util.Stack;
import javassist.CannotCompileException;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.NotFoundException;
import leodbc.javaagent.res.Constants;
import leodbc.javaagent.transformations.TransformerUtils;

/**
 * Rappresenta un metodo generato per la gestione di un contratto
 * 
 * @author 2638966
 */
public abstract class GeneratedMethod 
{
	/**
	 * Metodo per il quale viene fatta la generazione
	 */
	protected CtMethod method;
	
	/**
	 * Generatore inizializzato per un certo metodo
	 * @param m
	 */
	public GeneratedMethod(CtMethod m)
	{
		method=m;
	}
	
	
	/**
	 * @return Prefisso da applicare ai metodi generati
	 */
	public abstract String getMethodPrefix();
	
	/**
	 * @return Nome del metodo da generare
	 */
	public String getGeneratedMethodName() throws NotFoundException
	{
		return getMethodPrefix()+method.getName();
	}
	
	/**
	 * @return Nome del metodo da generare con nome di classe esplicito
	 */
	public String getGeneratedMethodNameExplicitClass(CtMethod met) throws NotFoundException
	{
		return getMethodPrefix()+met.getName()+"_"+met.getDeclaringClass().getName().replace('.', '_');
	}

	/**
	 * Gestisce un tag trovato in una interfaccia e ritorna
	 * la stringa del codice da generare per la sua gestione
	 * @param tag Tag trovato (il codice deve verificarne il tipo)
	 * @param clazz Classe che implemententa l'interfaccia
	 * @param method Metodo dove c' l'annotazione
	 * @return Codice di gestione
	 */
	public abstract String handleTagInInterface(CtClass clazz, CtMethod method, Object tag) throws CannotCompileException, ClassNotFoundException, NotFoundException;	

	/**
	 * @return Schema del metodo da generare
	 */
	public abstract String getMethodSchema(); 
	

	/**
	 * Tipologie dei parametri del metodo generato
	 * @param method Metodo da usare per la generazione
	 * @return Come da descrizione
	 * @throws NotFoundException
	 */
	public abstract CtClass[] generatedMethodParameterTypes() throws NotFoundException;
	
	/**
	 * @return Genera l'invocazione al metodo della classe padre
	 */
	public abstract String getSuperclassMethodInvocation() throws NotFoundException;
	
	/**
	 * Inserisce nel metodo la chiamata al nuovo metodo generato
	 * @throws CannotCompileException
	 */
	public abstract void injectMethodCall() throws NotFoundException, CannotCompileException;	

	/**
	 * Gestione dei contratti nelle interfacce
	 * @param method Metodo
	 * @return Operazioni da fare per garantire il contratto delle interfacce
	 */
	public String handleInterfaces() throws CannotCompileException, ClassNotFoundException, NotFoundException {
		String result="";
		Stack<CtClass> interfacesStack=new Stack<CtClass>();
		CtClass clazz=method.getDeclaringClass();
		for(CtClass inter: clazz.getInterfaces())
		{
			interfacesStack.push(inter);
		}
		
		while(!interfacesStack.empty())
		{
			CtClass inter=interfacesStack.pop();
			
			// Elaborazione della interfaccia
			for(CtMethod intMethod: inter.getDeclaredMethods())
			{
				if(intMethod.getSignature().equals(method.getSignature()))
				{
					for(Object an: intMethod.getAnnotations())
					{
						String tmp=handleTagInInterface(clazz, intMethod, an);
						if(tmp.length()!=0)
						{
							result+=tmp;
						}
					}
				}
			}
			
			// Elaborazione della sua superclasse
			for(CtClass subinter: inter.getInterfaces())
			{
				interfacesStack.push(subinter);
			}
		}
		return result;
	}

	/**
	 * Gestione di un tag
	 * @param ld Tag
	 * @return Nome del metodo di precondizione generato
	 */
	public void handleTag(Object ld) throws CannotCompileException, ClassNotFoundException, NotFoundException 
	{
		CtClass clazz=method.getDeclaringClass();

		// Parametri come tipologie
		CtClass[] methodTypes = generatedMethodParameterTypes();
		
		// Controllo della superclasse
		String preconditionMethodName=getGeneratedMethodName();
		String superMethodInvocation="";
		
		// Method parameters
		if(TransformerUtils.classHasMethod(clazz.getSuperclass(), preconditionMethodName, methodTypes))
		{
			superMethodInvocation=getSuperclassMethodInvocation();
		}
		superMethodInvocation+=handleInterfaces();

		if(ld!=null || superMethodInvocation.trim().length()!=0)
		{
			// Scrittura del metodo delle precondizioni
			String methodCode = generateMethodCode(method, ld, superMethodInvocation);
			methodCode="public void "+getGeneratedMethodName()+"("+generatedMethodParameterList()+") "+methodCode;
			
			if(Constants.INTERNAL_DEBUGGING)
			{
				System.out.println(clazz.getName()+": "+methodCode);
			}
			
			// Aggiunta del metodo delle precondizioni
			CtMethod preMethod=CtNewMethod.make(methodCode, clazz);
			clazz.addMethod(preMethod);
			injectMethodCall();
		}
	}	
	
	/**
	 * Lista dei parametri generato per un metodo
	 * @param method Metodo dal quale estrarre la lista dei parametri
	 * @return Lista dei parametri
	 * @throws NotFoundException
	 */
	public abstract String generatedMethodParameterList() throws NotFoundException;
	
	/**
	 * Generazine del codice del metodo
	 * @param ld Tag
	 * @param superMethodInvocation
	 * @param assertions
	 * @return
	 * @throws NotFoundException
	 */
	public String generateMethodCode(CtBehavior method, Object ld, String superMethodInvocation) throws NotFoundException {
		String assertions = TransformerUtils.extractValueFromAnnotation(ld);
		assertions=assertions.trim().length()==0?"true":assertions;
		String methodCode=TransformerUtils.replace(getMethodSchema(), Constants.CONDITION_MARKER, assertions, true);
		methodCode=TransformerUtils.replace(methodCode, Constants.LAUNCH_SUPERCLASS_MARKER, superMethodInvocation, true);
		return methodCode;
	}


	/**
	 * Gestione di un tag di precondizione su di una interfaccia
	 * @param clazz Classe nella quale aggiungere il metodo
	 * @param method Metodo
	 * @param ld Tag
	 * @return Nome del metodo
	 */
	public String generateMethodFromInterfaceTag(CtClass clazz, Object ld, CtMethod met) throws CannotCompileException, ClassNotFoundException, NotFoundException {
		String methodParams = generatedMethodParameterList();
		
		// Controllo della superclasse
		// Scrittura del metodo delle precondizioni
		String preconditionMethodName=getGeneratedMethodNameExplicitClass(met);
		String methodCode=generateMethodCode(met, ld, "");
		methodCode="public void "+preconditionMethodName+"("+methodParams+") "+methodCode;
		
		if(Constants.INTERNAL_DEBUGGING)
		{
			System.out.println(clazz.getName()+": "+methodCode);
		}
		
		// Aggiunta del metodo delle precondizioni
		CtMethod preMethod=CtNewMethod.make(methodCode, clazz);
		clazz.addMethod(preMethod);		
		return preconditionMethodName;
	}	
}
