package leodbc.javaagent.instrumentor;

import java.io.IOException;
import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.IllegalClassFormatException;
import java.security.ProtectionDomain;
import java.util.LinkedList;
import java.util.List;

import javassist.CannotCompileException;
import javassist.CtClass;
import javassist.NotFoundException;
import leodbc.javaagent.transformations.ArgCheckTransformer;
import leodbc.javaagent.transformations.ClassTrasformerList;
import leodbc.javaagent.transformations.InvariantTransformer;
import leodbc.javaagent.transformations.PostconditionTransformer;
import leodbc.javaagent.transformations.PreconditionTransformer;

/**
 * Trasformatore dei file class secondo le specifiche del DBC
 * 
 * @author Leonardo Cecchi
 */
public class DbcClassFileTransformer implements ClassFileTransformer 
{
	/**
	 * Trasformatore
	 */
	private ClassTrasformerList trasformerList;
	
	/**
	 * Mapping fra classloader e classpool
	 */
	private ClassLoaderToClassPool mapping;
	
	/**
	 * Cache di trasformazione
	 */
	private InstrumentorCache cache;
	
	/**
	 * Inizializzazione del trasformatore
	 */
	public DbcClassFileTransformer()
	{
		mapping=new ClassLoaderToClassPool();
		cache=new InstrumentorCache();
		trasformerList=new ClassTrasformerList();
		trasformerList.addClassTransformer(new PreconditionTransformer());
		trasformerList.addClassTransformer(new PostconditionTransformer());
		trasformerList.addClassTransformer(new InvariantTransformer());
		trasformerList.addClassTransformer(new ArgCheckTransformer());
	}
	
	/**
	 * Trasformazione di una singola classe
	 */
	public synchronized byte[] transform(ClassLoader loader, String className, Class<?> classBeingRedefined, ProtectionDomain protectionDomain, byte[] classfileBuffer) throws IllegalClassFormatException 
	{
		String dottedClassName=className.replace('/', '.');
		
		try
		{
			CtClass clazz=mapping.getClassPoolForClassLoader(loader).get(dottedClassName);
			
			if(clazz.isInterface())
			{
				return null;
			}
			
			List<CtClass> superClasses=getSuperClasses(clazz);
			for(CtClass thisClazz: superClasses)
			{
				if(cache.getFromCache(thisClazz.getName())==null)
				{
					cache.setCache(thisClazz.getName(), instrumentByteCode(thisClazz));
				}
			}
			return cache.getFromCache(clazz.getName());
		} catch(NotFoundException nfe)
		{
			nfe.printStackTrace();
			return null;
		}
	}

	/**
	 * Ritorna una lista contenente le superclassi della classe passata e 
	 * la classe passata in ordine dalla superclasse verso la sottoclasse
	 * @param clazz Classe
	 * @return Come sopra
	 */
	public List<CtClass> getSuperClasses(CtClass clazz) throws NotFoundException
	{
		CtClass thisClazz=clazz;
		LinkedList<CtClass> list=new LinkedList<CtClass>();
		
		while(thisClazz!=null && !thisClazz.getName().equals("java.lang.Object"))
		{
			list.addFirst(thisClazz);
			thisClazz=thisClazz.getSuperclass();
		}
		
		return list;
	}
	
	/**
	 * Modifica un bytecode per aderire ai contratti
	 * 
	 * @param clazz CtClass da modificare
	 * @return Bytecode modificato oppure null se tutto rimane uguale
	 */
	public byte[] instrumentByteCode(CtClass clazz)
	{
		// Ok!
		try
		{
			trasformerList.transform(clazz);
			byte[] thisByteCode=clazz.toBytecode();
			return thisByteCode;
		} catch(IOException ioe)
		{
			ioe.printStackTrace();
			return null;
		} catch(CannotCompileException cce)
		{
			System.err.println("Compilation error in "+clazz.getName().toString()+": "+cce.getReason());
			return null;
		} catch(ClassNotFoundException cnfe)
		{
			System.err.println(cnfe.toString());
			cnfe.printStackTrace();
			return null;
		} catch(NotFoundException nfe)
		{
			System.err.println(nfe.toString());
			nfe.printStackTrace();
			return null;
		}
	}
}
